#include <errno.h>
#include <unistd.h>
#include <ioctl-types.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>



/*

br (BottleRocket) (c) 1999 Tymm Twillman (tymm@acm.org)

Free controller software thingie for the X10(R) FireCracker wireless computer
interface kit.

This is distributed under the LGPL license.  It's not guaranteed to be useful,
work, run, compile, or anything else.

Have fun with it and if you do anything really cool, send an email and let me
know.


*/




/*
 * Yeah.  I know.  It'll get moved to ttySx soon enough... Maybe some day if 
 *   you're really lucky, I'll even add a command-line option.  Woo!
 */

#define X10_DEVICE "/dev/cua0"

#define ON 0
#define OFF 1
#define DIM 4
#define BRIGHT 12 /* upper bit gets masked off; just to tell dim/bright apart */

/*
 * If you have problems, try increasing BIT_WAIT and DELAY
 */

#define BIT_WAIT 3
#define DELAY 100000

/*
 * Used to create letter group portion of command -- could use some bit
 *  magic but this is less of a pain and easier to read
 */

char group_table[16] = {
  /* A */ 0x06, /* B */ 0x07, /* C */ 0x04, /* D */ 0x05,
  /* E */ 0x08, /* F */ 0x09, /* G */ 0x0a, /* H */ 0x0b,
  /* I */ 0x0e, /* J */ 0x0f, /* K */ 0x0c, /* L */ 0x0d,
  /* M */ 0x00, /* N */ 0x01, /* O */ 0x02, /* P */ 0x03
};

char *MyName = "bottlerocket";   /* Default */

void delay()
{
	/*
	 * Before and after a command, we have to wait a bit with DTR/RTS
	 *  set...
	 */

	struct timeval tv = { 0, DELAY };

	/*
	 * Not expecting any signals, so this should be safe
	 */

	if (select(0, NULL, NULL, NULL, &tv) < 0) {
		fprintf(stderr, "%s:  Error %d on select.\n", MyName, errno);
		perror(MyName);
		exit(errno);
	} 
}

void bits_out(const int fd, const int bits)
{
	/*
	 * Send out one command bit; set RTS or DTR (but only one) depending on
	 *  value.
	 */

	int loop;
	int out;
	int wait;


	out = (bits) ? TIOCM_RTS:TIOCM_DTR;

#ifdef DEBUG
	/*
	 * I print these things out funny because that's how I started doing
	 *  it and so too bad if it looks weird.
	 */

	printf("%d", out >> 1);
#endif

	/* Set RTS, DTR to desired settings */

	if (ioctl(fd, TIOCMSET, &out) < 0) {
		fprintf(stderr, "%s:  Error %d on ioctl in bits_out.\n",
		  MyName, errno);
		perror(MyName);
		exit(errno);
	}

	/*
	 * Junk just to make sure that we don't talk faster than it can listen
	 *  -- most delays have at best 1/100 second resolution on a PC
	 *  (they ask for preemption); so instead we use port
	 *  access overhead.  Not pretty, but seems to work...
	 */

	for (loop = 0; loop < BIT_WAIT; loop++) 
		ioctl(fd, TIOCMGET, &out);
}

void clock_out(const int fd)
{
	/*
	 * Send out a "clock pulse" -- both RTS and DTR set; used before/after
	 *  command (long pulse) and between command bits (short)
	 */

	int loop;
	int out = TIOCM_RTS | TIOCM_DTR;

#ifdef DEBUG
	printf("%d", out >> 1);
#endif

	if (ioctl(fd, TIOCMSET, &out) < 0) {
		fprintf(stderr, "%s:  Error %s on ioctl in clock_out.\n",
		  MyName, errno);
		perror(MyName);
		exit(errno);
	}

	/*
	 * More delay junk.  See above.
	 */

	for (loop = 0; loop < BIT_WAIT; loop++)
		ioctl(fd, TIOCMGET, &out);
}

void usage()
{
	fprintf(stderr, "BottleRocket version 0.01                      \n");
	fprintf(stderr, "(c) 1999 - Tymm Twillman                       \n");
	fprintf(stderr, "                                               \n");
	fprintf(stderr, "Purpose:  Control X10 \"Dynamite\" wireless    \n");
	fprintf(stderr, "           home-control interface things.      \n");
	fprintf(stderr, "                                               \n");
	fprintf(stderr, "Usage: %s <A-P><1-16> (on | off)               \n",
	  MyName);
	fprintf(stderr, "       %s <A-P> (bright | dim)                 \n",
	  MyName);
	fprintf(stderr, "                                               \n");
	fprintf(stderr, "Examples: %s A1 on                             \n",
	  MyName);
	fprintf(stderr, "          %s P8 off                            \n",
	  MyName);
	fprintf(stderr, "          %s B dim                             \n",
	  MyName);
}

void cmd_out(int fd, int group, int device, int cmd)
{

	/*
	 * Put together the commands to send out.  The basic start and end of 
	 *  each command is the same; just fill in the little bits in the middle
	 *
	 * Yeah, it's pretty nasty; I'll probably get around to cleaning this
	 *  up soon, but until then you just have to suffer.
	 */

	unsigned char cmd_seq[5] = { 0xd5, 0xaa, 0x00, 0x00, 0xad };

	register int i;
	register int j;
	unsigned char byte;
 	int out;


	/*
	 * Slap together the variable part of a command
	 */

	cmd_seq[2] |= (group << 4) | ((device & 0x08) >> 1);
	cmd_seq[3] |= ((cmd & 0x05) << 5) | ((device & 0x04) << 4) | ((device & 0x01) << 4)
	  | ((device & 0x02) << 2) | (((cmd == BRIGHT) | (cmd == DIM)) << 3)
	  | ((cmd == DIM) << 4);;

#ifdef DEBUG
	printf("cmd: %d\n", cmd);
#endif

	for (j = 0; j < 5; j++) {
		byte = cmd_seq[j];

#ifdef UGLY_DEBUG
		printf("sending byte: %02x\n", (unsigned int)byte);
#endif

		/*
		 * Roll out the bits, following each one by a "clock".
		 */

		for (i = 0; i < 8; i++) {
            		out = (byte & 0x80) ? 1:0;
			byte <<= 1;
			bits_out(fd, out);
			clock_out(fd);
		}
	}
}

char get_cmd(char *arg)
{
	if (!strcasecmp("on", arg)) {
		return ON;
	} else if (!strcasecmp("off", arg)) {
		return OFF;
	} else if (!strcasecmp("dim", arg)) {
		return DIM;
	} else if (!strcasecmp("bright", arg)) {
		return BRIGHT;
	}

	return -1;
}

char get_device(char *arg, char cmd)
{
	char device;


	/*
	 * Figure out the numeric part of the address of a unit (which device
	 *  in a group) we want to talk to
	 *
	 * If the command is a "dim" command, the device should be set to 0.
	 *  Unfortunately, some of the address lines are used for the dim/
	 *  bright commands.
	 */

	if ((cmd == DIM) || (cmd == BRIGHT)) {
                if (strlen(arg) != 1)
			return -1;
		device = 0;
	} else {

		if (strlen(arg) < 2)
			return -1;
		device = (char)atoi(&arg[1]) - 1;
		if ((device < 0) || (device > 15))
			return -1;
	}

	return device;
}

char get_group(char *arg)
{
	/*
	 * Get the letter part of a unit address (what "group" of units 
	 *  requested one is in)
	 */

	int tmp;


	if (!strlen(arg))
		return -1;

	tmp = toupper(arg[0]) - 'A';
	if ((tmp > 15) || (tmp < 0))
		return -1;

	/*
	 * Gotta look up the representation of the group that these things
	 *   use...
	 */

	return group_table[tmp];
}
	
int main(int argc, char *argv[])
{
	char device;
	char group;
	char cmd;
 	int fd;

	/*
	 * Who am I?
	 */

	if (argc > 0)
		MyName = argv[0];

	/*
	 * What am I doing?
	 */
	
	if ((argc < 3)
	  || ((cmd = get_cmd(argv[2])) < 0)
	  || ((device = get_device(argv[1], cmd)) < 0)
	  || ((group = get_group(argv[1])) < 0)
	)
	{
		usage();
		exit(0);
	}

	/*
	 * Snag the serial port...
	 */

	if ((fd = open(X10_DEVICE, O_RDWR)) < 0) {
		fprintf(stderr, "%s:  Error %d opening %s.\n", MyName, errno,
		  X10_DEVICE);
		perror(MyName);
		exit(errno);
	}

#ifdef DEBUG
	fprintf(stderr, "%s:  Opened %s.\n", MyName, X10_DEVICE);
	fflush(stdout);
#endif

	/*
	 * Send a "clock" pulse and wait for a bit to let the receiver know
	 *  it should pay attention
	 */

	clock_out(fd);
	delay();

	/*
	 * Do the deed...
	 */

	cmd_out(fd, group, device, cmd);

	/*
	 * Make sure lines stay in clock position for a bit
	 *  to allow command to complete
	 */

	delay();

	exit(0);
}
